import copy
from numbers import Number
from typing import Dict, Set, Tuple

from .grammars import GRAMMARS

from .cell_indices import CellIndices
from .cell_type import CellType

# ==== TYPES ====
# Type for the word of some language, logic: key: language, value: word
T_word = Dict[str, str]
# ===============


class WordConstructor(object):
    """Provides functionality for constructing words in all supported languages
        and also serves container for keeping them.

    Attributes:
        languages (Set[str]): What languages are used.
        words (Dict[str, str]): Mapping from language name to word.
    """

    def __init__(self, *,
                 words: T_word = None,
                 languages: Set[str] = None,
                 cell_indices: CellIndices
                 ):
        """Initialise the word cunstructor.

        Args:
            languages (Set[str]): What languages are used.
            words (Dict[str, str]): Mapping from language name to word.
        """
        if languages is not None:
            self.languages: Set[str] = languages
        else:
            self.languages: Set[str] = set(cell_indices.languages)

        if words is not None:
            self.words: T_word = words
        else:
            self.words: T_word = {key: "" for key in self.languages}

    @staticmethod
    def init_from_new_cell(cell, /) -> 'WordConstructor':  # noqa E999
        """Initialise the words when the new cell is created.

        Args:
            cell (Cell): just created cell.

        Returns:
            WordConstructor: words for the new cell.
        """
        if cell.value is not None:
            # Value type cell
            return WordConstructor.constant(cell)
        else:
            # Empty cell
            return WordConstructor(cell_indices=cell.cell_indices)

    @staticmethod
    def _binary_operation(first,
                          second,
                          operation: str) -> 'WordConstructor':
        """General binary operation.

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.
            operation (str): Definition of the operation (in grammar).

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        instance = copy.deepcopy(first.word)
        first_words = first.word.words
        second_word = second.word.words
        for language in instance.languages:
            pref = GRAMMARS[language]['operations'][operation]['prefix']
            suff = GRAMMARS[language]['operations'][operation]['suffix']
            sp = GRAMMARS[language]['operations'][operation]['separator']
            instance.words[language] = pref + first_words[language] + sp \
                + second_word[language] + suff
        return instance

    @staticmethod
    def add(first, second, /) -> 'WordConstructor':  # noqa E225
        """Add binary operation (+).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "add")

    @staticmethod
    def subtract(first, second, /) -> 'WordConstructor':  # noqa E225
        """Subtract binary operation (-).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "subtract")

    @staticmethod
    def multiply(first, second, /) -> 'WordConstructor':  # noqa E225
        """Multiply binary operation (*).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "multiply")

    @staticmethod
    def divide(first, second, /) -> 'WordConstructor':  # noqa E225
        """Divide binary operation (/).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "divide")

    @staticmethod
    def modulo(first, second, /) -> 'WordConstructor':  # noqa E225
        """Modulo binary operation (%).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "modulo")

    @staticmethod
    def power(first, second, /) -> 'WordConstructor':  # noqa E225
        """Power binary operation (**).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "power")

    @staticmethod
    def logicalConjunction(first,
                           second, /) -> 'WordConstructor':  # noqa E225
        """Logical conjunction operation (and).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second,
                                                 "logical-conjunction")

    @staticmethod
    def logicalDisjunction(first,
                           second, /) -> 'WordConstructor':  # noqa E225
        """Logical disjunction operation (or).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second,
                                                 "logical-disjunction")

    @staticmethod
    def equalTo(first, second, /) -> 'WordConstructor':  # noqa E225
        """Equal to binary operation (==).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "equal-to")

    @staticmethod
    def notEqualTo(first, second, /) -> 'WordConstructor':  # noqa E225
        """Not equal to binary operation (!=).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "not-equal-to")

    @staticmethod
    def greaterThan(first,
                    second, /) -> 'WordConstructor':  # noqa E225
        """Greater than binary operation (>).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "greater-than")

    @staticmethod
    def greaterThanOrEqualTo(first,
                             second, /) -> 'WordConstructor':  # noqa E225
        """Greater than or equal to binary operation (>=).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second,
                                                 "greater-than-or-equal-to")

    @staticmethod
    def lessThan(first, second, /) -> 'WordConstructor':  # noqa E225
        """Less than binary operation (<).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second, "less-than")

    @staticmethod
    def lessThanOrEqualTo(first,
                          second, /) -> 'WordConstructor':  # noqa E225
        """Less than or equal to binary operation (<=).

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator and two
                operands.
        """
        return WordConstructor._binary_operation(first, second,
                                                 "less-than-or-equal-to")

    @staticmethod
    def concatenate(first,
                    second) -> 'WordConstructor':
        """Concatenate values as two strings and create a word.

        Args:
            first (Cell): The first cell (operand) of the operator.
            second (Cell): The second cell (operand) of the operator.

        Returns:
            WordConstructor: Word constructed using binary operator concatenate
                and two operands.
        """
        # Inline function for appending context to inputs
        def _generate_word_string_concatenate(in_cell) -> T_word:
            """Generate word for each operand.

            Returns:
                T_word: Words for each language.
            """
            if in_cell.anchored:
                return WordConstructor.reference(in_cell).words
            else:
                if in_cell.cell_type == CellType.computational:
                    return in_cell._constructing_words.words
                elif in_cell.cell_type == CellType.value_only:
                    words: T_word = {}
                    for language in instance.languages:
                        if isinstance(in_cell.value, str):
                            pref = GRAMMARS[language]['operations'][
                                'concatenate']['string-value']['prefix']
                            suff = GRAMMARS[language]['operations'][
                                'concatenate']['string-value']['suffix']
                            words[language] = pref + str(in_cell.value) + suff
                        if isinstance(in_cell.value, Number):
                            pref = GRAMMARS[language]['operations'][
                                'concatenate']['numeric-value']['prefix']
                            suff = GRAMMARS[language]['operations'][
                                'concatenate']['numeric-value']['suffix']
                            words[language] = pref + str(in_cell.value) + suff
                    return words

        instance = copy.deepcopy(first.word)
        first_words = _generate_word_string_concatenate(first)
        second_word = _generate_word_string_concatenate(second)
        for language in instance.languages:
            pref = GRAMMARS[language]['operations']['concatenate']['prefix']
            suff = GRAMMARS[language]['operations']['concatenate']['suffix']
            sp = GRAMMARS[language]['operations']['concatenate']['separator']
            instance.words[language] = pref + first_words[language] + sp \
                + second_word[language] + suff
        return instance

    @staticmethod
    def _aggregation_parse_cell(cell,
                                start_idx: Tuple[int, int],
                                end_idx: Tuple[int, int],
                                language: str
                                ) -> Tuple[str, str]:
        """Generates the aggregation string for a concrete language (regardless
            what aggregation method is selected).

        Args:
            cell (Cell): Any cell in the set (used for extracting indicies).
            start_idx (Tuple[int, int]): Position of the slice start.
            end_idx (Tuple[int, int]): Position of the slice end.
            language (str): What language is used.

        Returns:
            Tuple[str, str]: Definition of starting, ending cell in language.
        """
        # Does the language include the last cell?
        #   if yes, offset of size 1 has to be included.
        offset = 0
        if GRAMMARS[
            language
        ]['cells']['aggregation']['include_last_cell']:
            offset = 1

        start_idx_r = cell.cell_indices.rows[language][start_idx[0]]
        start_idx_c = cell.cell_indices.columns[language][start_idx[1]]
        end_idx_r = cell.cell_indices.rows[language][end_idx[0] + offset]
        end_idx_c = cell.cell_indices.columns[language][end_idx[1] + offset]

        pref_cell_start = GRAMMARS[language]['cells']['aggregation'][
            'start_cell']['prefix']
        separator_cell_start = GRAMMARS[language]['cells']['aggregation'][
            'start_cell']['separator']
        suffix_cell_start = GRAMMARS[language]['cells']['aggregation'][
            'start_cell']['suffix']

        if GRAMMARS[language]['cells']['aggregation']['start_cell'][
            'row_first'
        ]:
            cell_start = (pref_cell_start + start_idx_r +
                          separator_cell_start + start_idx_c +
                          suffix_cell_start)
        else:
            cell_start = (pref_cell_start + start_idx_c +
                          separator_cell_start + start_idx_r +
                          suffix_cell_start)

        if GRAMMARS[language]['cells']['aggregation']['start_cell'][
            'rows_only'
        ]:
            cell_start = (pref_cell_start + start_idx_r +
                          separator_cell_start + end_idx_r +
                          suffix_cell_start)
        elif GRAMMARS[language]['cells']['aggregation']['start_cell'][
            'cols_only'
        ]:
            cell_start = (pref_cell_start + start_idx_c +
                          separator_cell_start + end_idx_c +
                          suffix_cell_start)
        # ---------- ending cell -----
        pref_cell_end = GRAMMARS[language]['cells']['aggregation'][
            'end_cell']['prefix']
        separator_cell_end = GRAMMARS[language]['cells']['aggregation'][
            'end_cell']['separator']
        suffix_cell_end = GRAMMARS[language]['cells']['aggregation'][
            'end_cell']['suffix']

        if GRAMMARS[language]['cells']['aggregation']['end_cell'][
            'row_first'
        ]:
            cell_end = (pref_cell_end + end_idx_r +
                        separator_cell_end + end_idx_c +
                        suffix_cell_end)
        else:
            cell_end = (pref_cell_end + end_idx_c +
                        separator_cell_end + end_idx_r +
                        suffix_cell_end)

        if GRAMMARS[language]['cells']['aggregation']['end_cell'][
            'rows_only'
        ]:
            cell_end = (pref_cell_end + start_idx_r +
                        separator_cell_end + end_idx_r +
                        suffix_cell_end)
        elif GRAMMARS[language]['cells']['aggregation']['end_cell'][
            'cols_only'
        ]:
            cell_end = (pref_cell_end + start_idx_c +
                        separator_cell_end + end_idx_c +
                        suffix_cell_end)
        # ----------------------------
        return cell_start, cell_end

    @staticmethod
    def aggregation(cell_start, cell_end,
                    grammar_method: str) -> 'WordConstructor':
        """General aggregation function.

        Args:
            cell_start (Cell): The cell on the position of the slice start.
            cell_end (Cell): The cell on the position of slice end.
            grammar_method (str): Name of the aggregation method in a grammar.

        Returns:
            WordConstructor: World constructed from aggregation method.
        """
        if not(cell_start.anchored and cell_end.anchored):
            raise ValueError("Both starting end ending cells has to be"
                             " anchored!")

        start_idx = cell_start.coordinates
        end_idx = cell_end.coordinates

        instance = WordConstructor(cell_indices=cell_start.cell_indices)
        for language in instance.languages:
            prefix = GRAMMARS[language]['cells']['aggregation']['prefix']
            separator = GRAMMARS[language]['cells']['aggregation']['separator']
            suffix = GRAMMARS[language]['cells']['aggregation']['suffix']

            start, end = WordConstructor._aggregation_parse_cell(
                cell_start, start_idx, end_idx, language
            )
            # Methods
            m_prefix = GRAMMARS[language]['operations'][grammar_method][
                'prefix'
            ]
            m_suffix = GRAMMARS[language]['operations'][grammar_method][
                'suffix'
            ]

            instance.words[language] = (m_prefix + prefix + start +
                                        separator + end + suffix + m_suffix)
        return instance

    @staticmethod
    def parse(cell) -> T_word:
        """Parse the cell word. This function is called when the cell should
            be inserted to spreadsheet.

        Args:
            cell (Cell): The cell that is the subject of parsing.

        Returns:
            T_word: Parsed cell.
        """
        if cell.cell_type == CellType.value_only:
            if cell.value is not None:
                # Constant value
                return copy.deepcopy(WordConstructor.constant(cell).words)
            # Empty value
            return copy.deepcopy(WordConstructor.empty(cell).words)
        elif cell.cell_type == CellType.computational:
            # Computational type
            words: T_word = copy.deepcopy(cell.constructing_words.words)
            for language in cell.constructing_words.languages:
                prefix = GRAMMARS[language]['cells']['operation']['prefix']
                suffix = GRAMMARS[language]['cells']['operation']['suffix']
                words[language] = prefix + words[language] + suffix
            return words

    @staticmethod
    def raw(cell, words: T_word, /) -> 'WordConstructor':  # noqa: E225
        """Returns the raw statement string.

        Args:
            cell (Cell): Some cell.
            words (T_word): Custom word definition

        Returns:
            WordConstructor: Defined word.
        """
        instance = WordConstructor(cell_indices=cell.cell_indices)
        for language in instance.languages:
            instance.words[language] = words[language]
        return instance

    @staticmethod
    def empty(cell, /) -> 'WordConstructor':  # noqa E225
        """Returns the empty string.

        Args:
            cell (Cell): Empty cell (without any value or computation)

        Returns:
            WordConstructor: Word with empty string.
        """
        instance = WordConstructor(cell_indices=cell.cell_indices)
        for language in instance.languages:
            content = GRAMMARS[language]['cells']['empty']['content']
            instance.words[language] = content
        return instance

    @staticmethod
    def reference(cell, /) -> 'WordConstructor':  # noqa E225
        """Return the reference to the cell.

        Args:
            cell (Cell): The cell to which value is referenced.

        Returns:
            WordConstructor: Word with reference
        """
        instance = WordConstructor(cell_indices=cell.cell_indices)
        for language in instance.languages:
            prefix = GRAMMARS[language]['cells']['reference']['prefix']
            separator = GRAMMARS[language]['cells']['reference']['separator']
            suffix = GRAMMARS[language]['cells']['reference']['suffix']
            row_first = GRAMMARS[language]['cells']['reference']['row_first']
            # Parse the position to the text of the column and row
            col_parsed = cell.cell_indices.columns[language][cell.column]
            row_parsed = cell.cell_indices.rows[language][cell.row]
            body = prefix + col_parsed + separator + row_parsed + suffix
            if row_first:
                body = prefix + row_parsed + separator + col_parsed + suffix
            instance.words[language] = body
        return instance

    @staticmethod
    def constant(cell, /) -> 'WordConstructor':  # noqa E225
        """Return the value of the cell.

        Args:
            cell (Cell): The cell which value is considered.

        Returns:
            WordConstructor: Word with value.
        """
        instance = WordConstructor(cell_indices=cell.cell_indices)
        for language in instance.languages:
            prefix = GRAMMARS[language]['cells']['constant']['prefix']
            suffix = GRAMMARS[language]['cells']['constant']['suffix']
            instance.words[language] = prefix + str(cell.value) + suffix
        return instance

    @staticmethod
    def variable(cell, /) -> 'WordConstructor':  # noqa E225
        """Return the cell as a variable.

        Args:
            cell (Cell): The cell which is considered as a variable.

        Returns:
            WordConstructor: Word with variable.
        """
        instance = WordConstructor(cell_indices=cell.cell_indices)
        for language in instance.languages:
            prefix = GRAMMARS[language]['cells']['variable']['prefix']
            suffix = GRAMMARS[language]['cells']['variable']['suffix']
            # Now add the value as a suffix if required
            value_word = ""
            value_grammar = GRAMMARS[language]['cells']['variable']['value']
            if value_grammar['include']:
                value_word += value_grammar['prefix']
                value_word += str(cell.value)
                value_word += value_grammar['suffix']
            # Construct the whole word
            instance.words[language] = (prefix + cell.variable_name
                                        + value_word + suffix)

        return instance

    @staticmethod
    def _unary_operator(*,
                        cell,
                        prefix_path: Tuple[str],
                        suffix_path: Tuple[str]) -> 'WordConstructor':
        """Word creation logic for a general unary operator.

        Args:
            cell (Cell): The cell that is the body of the unary operator.
            prefix_path (Tuple[str]): The path to prefix of the unary operator.
            suffix_path (Tuple[str]): The path to suffix of the unary operator.

        Returns:
            'WordConstructor': Word constructed by the operator.
        """
        instance = copy.deepcopy(cell.word)
        for language in instance.languages:
            prefix = GRAMMARS[language]
            for path_item in prefix_path:
                prefix = prefix[path_item]
            suffix = GRAMMARS[language]
            for path_item in suffix_path:
                suffix = suffix[path_item]
            body = instance.words[language]
            instance.words[language] = prefix + body + suffix
        return instance

    @staticmethod
    def brackets(cell, /) -> 'WordConstructor':  # noqa E225
        """Add brackets around the cell.

        Args:
            cell (Cell): The cell around which brackets are added.

        Returns:
            WordConstructor: Word with brackets.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['brackets', 'prefix'],
            suffix_path=['brackets', 'suffix']
        )

    @staticmethod
    def logarithm(cell, /) -> 'WordConstructor':  # noqa E225
        """Add logarithm definition context around the cell.

        Args:
            cell (Cell): The cell around which logarithm context is added.

        Returns:
            WordConstructor: Word with logarithm context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'logarithm', 'prefix'],
            suffix_path=['operations', 'logarithm', 'suffix']
        )

    @staticmethod
    def exponential(cell, /) -> 'WordConstructor':  # noqa E225
        """Add exponential definition context around the cell.

        Args:
            cell (Cell): The cell around which exponential context is added.

        Returns:
            WordConstructor: Word with exponential context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'exponential', 'prefix'],
            suffix_path=['operations', 'exponential', 'suffix']
        )

    @staticmethod
    def ceil(cell, /) -> 'WordConstructor':  # noqa E225
        """Add ceiling function definition context around the cell.

        Args:
            cell (Cell): The cell around which ceiling fn. context is added.

        Returns:
            WordConstructor: Word with ceiling function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'ceil', 'prefix'],
            suffix_path=['operations', 'ceil', 'suffix']
        )

    @staticmethod
    def floor(cell, /) -> 'WordConstructor':  # noqa E225
        """Add floor function definition context around the cell.

        Args:
            cell (Cell): The cell around which floor fn. context is added.

        Returns:
            WordConstructor: Word with floor function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'floor', 'prefix'],
            suffix_path=['operations', 'floor', 'suffix']
        )

    @staticmethod
    def round(cell, /) -> 'WordConstructor':  # noqa E225
        """Add rounding function definition context around the cell.

        Args:
            cell (Cell): The cell around which rounding fn. context is added.

        Returns:
            WordConstructor: Word with rounding function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'round', 'prefix'],
            suffix_path=['operations', 'round', 'suffix']
        )

    @staticmethod
    def abs(cell, /) -> 'WordConstructor':  # noqa E225
        """Add absolute value definition context around the cell.

        Args:
            cell (Cell): The cell around which absolute value context is added.

        Returns:
            WordConstructor: Word with absolute value function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'abs', 'prefix'],
            suffix_path=['operations', 'abs', 'suffix']
        )

    @staticmethod
    def sqrt(cell, /) -> 'WordConstructor':  # noqa E225
        """Add square root value definition context around the cell.

        Args:
            cell (Cell): The cell around which square root context is added.

        Returns:
            WordConstructor: Word with square root function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'sqrt', 'prefix'],
            suffix_path=['operations', 'sqrt', 'suffix']
        )

    @staticmethod
    def signum(cell, /) -> 'WordConstructor':  # noqa E225
        """Add signum value definition context around the cell.

        Args:
            cell (Cell): The cell around which signum context is added.

        Returns:
            WordConstructor: Word with signum function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'signum', 'prefix'],
            suffix_path=['operations', 'signum', 'suffix']
        )

    @staticmethod
    def logicalNegation(cell, /) -> 'WordConstructor':  # noqa E225
        """Add logical negation definition context around the cell.

        Args:
            cell (Cell): The cell around which logical negation context is
                added.

        Returns:
            WordConstructor: Word with logical negation function context.
        """
        return WordConstructor._unary_operator(
            cell=cell,
            prefix_path=['operations', 'logical-negation', 'prefix'],
            suffix_path=['operations', 'logical-negation', 'suffix']
        )

    @staticmethod
    def conditional(condition,
                    consequent,
                    alternative, /) -> 'WordConstructor':  # noqa E225
        """Construct the word for the conditional statement (if-then-else).

        Args:
            condition (Cell): condition statement.
            consequent (Cell): value if the condition is true.
            alternative (Cell): value if the condition is false.

        Returns:
            WordConstructor: Word for the conditional statement
        """
        instance = WordConstructor(cell_indices=condition.cell_indices)

        w_condition = condition.word.words
        w_consequent = consequent.word.words
        w_alternative = alternative.word.words

        for language in instance.languages:
            conditional_prefix = GRAMMARS[language]['conditional']['prefix']
            conditional_suffix = GRAMMARS[language]['conditional']['suffix']

            condition_prefix = GRAMMARS[language][
                'conditional']['condition']['prefix']
            condition_suffix = GRAMMARS[language][
                'conditional']['condition']['suffix']
            condition_word = condition_prefix + w_condition[language] + \
                condition_suffix

            consequent_prefix = GRAMMARS[language][
                'conditional']['consequent']['prefix']
            consequent_suffix = GRAMMARS[language][
                'conditional']['consequent']['suffix']
            consequent_word = consequent_prefix + w_consequent[language] + \
                consequent_suffix

            alternative_prefix = GRAMMARS[language][
                'conditional']['alternative']['prefix']
            alternative_suffix = GRAMMARS[language][
                'conditional']['alternative']['suffix']
            alternative_word = alternative_prefix + w_alternative[language] + \
                alternative_suffix

            # Swap the words to desired sequence
            clausules_order = GRAMMARS[language]['conditional']['order']
            prefered_order = ('condition', 'consequent', 'alternative')
            words_in_prefered_order = (condition_word, consequent_word,
                                       alternative_word)
            words_in_correct_order = ["", "", ""]
            # Do the permutation and construct the word:
            for clausule_idx, clausule in enumerate(prefered_order):
                clausule_permutated_idx = clausules_order.index(clausule)
                words_in_correct_order[clausule_permutated_idx] = \
                    words_in_prefered_order[clausule_idx]

            # Merge words into one word
            final_word = "".join(words_in_correct_order)

            instance.words[language] = conditional_prefix + final_word + \
                conditional_suffix

        return instance

    @staticmethod
    def offset(reference,
               row_skip,
               column_skip, /) -> 'WordConstructor':  # noqa E225
        """Construct the word for the offset statement (skip n rows and m
            columns from the referential position).

        Args:
            reference (Cell): Reference cell from that the position is
                computed.
            row_skip (Cell): How many rows (down) should be skipped.
            column_skip (Cell): How many columns (left) should be skipped.

        Returns:
            WordConstructor: Word for the offset statement.
        """
        instance = WordConstructor(cell_indices=reference.cell_indices)

        # For shorter access to rows and column indices:
        index_row = reference.cell_indices.rows
        index_col = reference.cell_indices.columns

        ref_row_skip = row_skip.word.words
        ref_col_skip = column_skip.word.words

        for language in instance.languages:
            offset_prefix = GRAMMARS[language]['cells']['offset']['prefix']
            offset_suffix = GRAMMARS[language]['cells']['offset']['suffix']

            ref_row_prefix = GRAMMARS[language]['cells']['offset'][
                'reference-cell-row']['prefix']
            ref_row_suffix = GRAMMARS[language]['cells']['offset'][
                'reference-cell-row']['suffix']
            ref_row_word = (ref_row_prefix +
                            index_row[language][reference.row] +
                            ref_row_suffix)

            ref_col_prefix = GRAMMARS[language]['cells']['offset'][
                'reference-cell-column']['prefix']
            ref_col_suffix = GRAMMARS[language]['cells']['offset'][
                'reference-cell-column']['suffix']
            ref_col_word = (ref_col_prefix +
                            index_col[language][reference.column] +
                            ref_col_suffix)
            # HERE USING REFERENCE
            skip_row_prefix = GRAMMARS[language]['cells']['offset'][
                'skip-of-rows']['prefix']
            skip_row_suffix = GRAMMARS[language]['cells']['offset'][
                'skip-of-rows']['suffix']
            skip_row_word = (skip_row_prefix +
                             ref_row_skip[language] +
                             skip_row_suffix)

            skip_col_prefix = GRAMMARS[language]['cells']['offset'][
                'skip-of-columns']['prefix']
            skip_col_suffix = GRAMMARS[language]['cells']['offset'][
                'skip-of-columns']['suffix']
            skip_col_word = (skip_col_prefix +
                             ref_col_skip[language] +
                             skip_col_suffix)

            # Swap the words to desired sequence
            clausules_order = GRAMMARS[language]['cells']['offset']['order']
            prefered_order = ('reference-cell-row', 'reference-cell-column',
                              'skip-of-rows', 'skip-of-columns')
            words_in_prefered_order = (ref_row_word, ref_col_word,
                                       skip_row_word, skip_col_word)
            words_in_correct_order = ["", "", "", ""]
            # Do the permutation and construct the word:
            for clausule_idx, clausule in enumerate(prefered_order):
                clausule_permutated_idx = clausules_order.index(clausule)
                words_in_correct_order[clausule_permutated_idx] = \
                    words_in_prefered_order[clausule_idx]

            # Merge words into one word
            final_word = "".join(words_in_correct_order)

            instance.words[language] = offset_prefix + final_word + \
                offset_suffix

        return instance



import abc
import json
from typing import Tuple, List, Dict, Union, Callable, Optional
from types import MappingProxyType
from numbers import Number

import xlsxwriter
import numpy

from .cell import Cell
from .cell_type import CellType
from .cell_indices import CellIndices

# ==== TYPES ====
# Type for the output dictionary with the logic:
# table->data->rows/columns->Row/Col label->columns/rows/help_text->
#   Column/Row label->cell keys (value, description, language alias)->value
T_out_dict = Dict[
    str,  # 'table', 'variables', 'row-labels', 'column-labels'
    Union[
        # Cell values:
        Dict[
            str,  # 'data'
            Dict[
                str,  # 'rows' exclusive or 'columns'
                Dict[
                    str,  # Row/Column key (label)
                    # For values:
                    Dict[
                        str,  # 'columns' xor 'rows' (or 'help_text')
                        Union[
                            Dict[
                                str,  # Column/Row key (label)
                                Dict[
                                    str,  # Cell key ('value', 'description',)
                                    Union[float, int, str]  # Cell values
                                    ]
                                ],
                            str  # For help text
                             ]
                        ]
                    ]
                ]
            ],
        # Variables:
        Dict[
            str,  # 'variables'
            Dict[
                str,  # variable name
                Dict[
                    str,  # 'value' or 'description'
                    Union[float, int, str]  # Variable value
                    ]
                ]
            ],
        # Labels of rows and columns
        Dict[
            str,  # 'row-labels' or 'column-labels'
            str  # Row/Column label
            ],
        # Optional dictionary values appended by user
        Optional[
            dict
            ]
        ],
    ]
# ===============


class Serialization(abc.ABC):
    """Provides basic functionality for exporting to required formats.

    Attributes:
        export_offset (Tuple[int, int]): Defines how many rows and
            columns are skiped from the left top corner. First index is
            number of rows, second number of columns.
        warning_logger (Callable[[str], None]): Function that logs the
            warnings.
        export_subset (bool): If true, warning are raised when exporting.
    """

    def __init__(self, *,
                 export_offset: Tuple[int, int] = (0, 0),
                 export_subset: bool = False,
                 warning_logger: Optional[Callable[[str], None]] = None):
        """Initialise functionality for serialization.

        Args:
            export_offset (Tuple[int, int]): Defines how many rows and
                columns are skiped from the left top corner. First index is
                number of rows, second number of columns.
            export_subset (bool): If true, warning are raised when exporting.
            warning_logger (Optional[Callable[[str], None]]): Function that
                logs the warnings (or None if skipped).
        """
        # Export offset of rows, columns
        self.export_offset: Tuple[int, int] = export_offset
        if warning_logger is not None:
            self.warning_logger: Callable[[str], None] = warning_logger
        else:
            # Silent logger
            self.warning_logger: Callable[[str], None] = lambda _mess: _mess
        self.export_subset: bool = export_subset

    @property
    def shape(self) -> Tuple[int, int]:
        """Get the shape as the tuple of number of rows and columns.

        Returns:
            Tuple[int, int]: number of rows, columns
        """
        raise NotImplementedError

    @property
    def cell_indices(self) -> CellIndices:
        """Get the cell indices.

        Returns:
            CellIndices: Cell indices of the spreadsheet.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _get_cell_at(self, row: int, column: int) -> Cell:
        """Get the particular cell on the (row, column) position.

        Returns:
            Cell: The call on given position.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _get_variables(self):
        """Return the sheet variables as _SheetVariables object.

        Returns:
            _SheetVariables: Sheet variables.
        """
        raise NotImplementedError

    def log_export_subset_warning_if_needed(self):
        """Log the export subset warning if needed.
        """
        if self.export_subset:
            self.warning_logger("Slice is being exported => there is"
                                " a possibility of data losses.")

    def to_excel(self,
                 file_path: str,
                 /, *,  # noqa: E225, E999
                 sheet_name: str = "Results",
                 spaces_replacement: str = ' ',
                 label_row_format: dict = {'bold': True},
                 label_column_format: dict = {'bold': True},
                 variables_sheet_name: Optional[str] = None,
                 variables_sheet_header: Dict[str, str] = MappingProxyType(
                     {
                         "name": "Name",
                         "value": "Value",
                         "description": "Description"
                     }),
                 values_only: bool = False
                 ) -> None:
        """Export the values inside Spreadsheet instance to the
            Excel 2010 compatible .xslx file

        Args:
            file_path (str): Path to the target .xlsx file.
            sheet_name (str): The name of the sheet inside the file.
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            label_row_format (dict): Excel styles for the label of rows,
                documentation: https://xlsxwriter.readthedocs.io/format.html
            label_column_format (dict): Excel styles for the label of columns,
                documentation: https://xlsxwriter.readthedocs.io/format.html
            variables_sheet_name (Optional[str]): If set, creates the new
                sheet with variables and their description and possibility
                to set them up (directly from the sheet).
            variables_sheet_header (Dict[str, str]): Define the labels (header)
                for the sheet with variables (first row in the sheet).
            values_only (bool): If true, only values (and not formulas) are
                exported.
        """
        # Quick sanity check
        if ".xlsx" not in file_path[-5:]:
            raise ValueError("Suffix of the file has to be '.xslx'!")
        if not isinstance(sheet_name, str) or len(sheet_name) < 1:
            raise ValueError("Sheet name has to be non-empty string!")
        # Log warning if needed
        self.log_export_subset_warning_if_needed()

        # Open or create an Excel file and create a sheet inside:
        workbook = xlsxwriter.Workbook(file_path)
        worksheet = workbook.add_worksheet(name=sheet_name)
        # Register the style for the labels:
        col_label_format = workbook.add_format(label_column_format)
        row_label_format = workbook.add_format(label_row_format)

        # Register all variables:
        if self._get_variables().empty:
            pass
        elif variables_sheet_name is None:
            for name, value in self._get_variables().variables_dict.items():
                workbook.define_name(name, str(value['value']))
        else:
            variables_sheet = workbook.add_worksheet(name=variables_sheet_name)
            # Insert header (labels)
            variables_sheet.write(0, 0, variables_sheet_header['name'],
                                  col_label_format)
            variables_sheet.write(0, 1, variables_sheet_header['value'],
                                  col_label_format)
            variables_sheet.write(0, 2, variables_sheet_header['description'],
                                  col_label_format)
            row_idx = 1
            for var_n, var_v in self._get_variables().variables_dict.items():
                # Insert variables to the sheet
                variables_sheet.write(row_idx, 0, var_n)
                variables_sheet.write(row_idx, 1, var_v['value'])
                variables_sheet.write(row_idx, 2, var_v['description'])
                # Register variable
                workbook.define_name(
                    var_n, f'={variables_sheet_name}!$B${row_idx + 1}'
                )
                row_idx += 1

        # Iterate through all columns and rows and add data
        for row_idx in range(self.shape[0]):
            for col_idx in range(self.shape[1]):
                cell: Cell = self._get_cell_at(row_idx, col_idx)
                if cell.value is not None:
                    # Offset here is either 0 or 1, indicates if we writes
                    # row/column labels to the first row and column.
                    offset_row = 0
                    offset_col = 0
                    if self.cell_indices.excel_append_row_labels:
                        offset_col = 1
                    if self.cell_indices.excel_append_column_labels:
                        offset_row = 1
                    # Excel format/style for the cell:
                    if len(cell.excel_format) > 0:
                        # Register the format
                        cell_format = workbook.add_format(cell.excel_format)
                    else:
                        cell_format = None
                    # Write actual data
                    if values_only or (cell.cell_type == CellType.value_only):
                        # If the cell is a value only, use method 'write'
                        worksheet.write(row_idx + offset_row,
                                        col_idx + offset_col,
                                        cell.value,
                                        cell_format)
                    else:
                        # If the cell is a formula, use method 'write_formula'
                        worksheet.write_formula(row_idx + offset_row,
                                                col_idx + offset_col,
                                                cell.parse['excel'],
                                                value=cell.value,
                                                cell_format=cell_format)
        # Add the labels for rows and columns
        if self.cell_indices.excel_append_column_labels:
            # Add labels of column
            for col_idx in range(self.shape[1]):
                worksheet.write(0,
                                col_idx + int(
                                    self.cell_indices.excel_append_row_labels
                                ),
                                self.cell_indices.columns_labels[
                                    # Reflect the export offset
                                    col_idx + self.export_offset[1]
                                ].replace(' ', spaces_replacement),
                                col_label_format)
        if self.cell_indices.excel_append_row_labels:
            # Add labels for rows
            for row_idx in range(self.shape[0]):
                worksheet.write(row_idx + int(
                    self.cell_indices.excel_append_column_labels
                ),
                                0,
                                self.cell_indices.rows_labels[
                                    # Reflect the export offset
                                    row_idx + self.export_offset[0]
                                ].replace(' ', spaces_replacement),
                                row_label_format)
        # Store results
        workbook.close()

    def to_dictionary(self,
                      languages: List[str] = None,
                      use_language_for_description: Optional[str] = None,
                      /, *,  # noqa E999
                      by_row: bool = True,
                      languages_pseudonyms: List[str] = None,
                      spaces_replacement: str = ' ',
                      skip_nan_cell: bool = False,
                      nan_replacement: object = None,
                      append_dict: dict = {},
                      generate_schema: bool = False
                      ) -> T_out_dict:
        """Export this spreadsheet to the dictionary.

        Args:
            languages (List[str]): List of languages that should be exported.
                If it has value None, all the languages are exported.
            use_language_for_description (Optional[str]): If set-up (using
                the language name), description field is set to be either
                the description value (if defined) or the value of this
                language.
            by_row (bool): If True, rows are the first indices and columns
                are the second in the order. If False it is vice-versa.
            languages_pseudonyms (List[str]): Rename languages to the strings
                inside this list.
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            skip_nan_cell (bool): If True, None (NaN) values are skipped.
            nan_replacement (object): Replacement for the None (NaN) value
            append_dict (dict): Append this dictionary to output.
            generate_schema (bool): If true, returns the JSON schema.

        Returns:
            dict:
                Output dictionary with the logic:
                table->data->rows/columns->Row/Col label->
                columns/rows/help_text->Column/Row label->
                cell keys (value, description, language alias)->value
        """
        if generate_schema:
            return self.generate_json_schema()

        # Log warning if needed
        self.log_export_subset_warning_if_needed()

        # Assign all languages if languages is None:
        if languages is None:
            languages = self.cell_indices.languages
        # Quick sanity check:
        if (
                languages_pseudonyms is not None
                and len(languages_pseudonyms) != len(languages)
        ):
            raise ValueError("Language pseudonyms does not have the same size "
                             "as the language array!")
        # Language array (use pseudonyms if possible, language otherwise)
        languages_used = languages
        if languages_pseudonyms is not None:
            languages_used = languages_pseudonyms
        # If by column (not by_row)

        # A) The x-axes represents the columns
        x_range = self.shape[1]
        x = [label.replace(' ', spaces_replacement)
             for label in self.cell_indices.columns_labels[
                          # Reflects the column offset for export
                          self.export_offset[1]:self.shape[1]
                          ]
             ]
        if (x_helptext := self.cell_indices.columns_help_text) is not None:  # noqa E203
            x_helptext_descriptor = "column_description"
            # Reflects the column offset for export
            x_helptext = x_helptext[self.export_offset[1]:self.shape[1]]
        x_start_key = 'columns'
        # The y-axes represents the rows
        y_range = self.shape[0]
        y = [label.replace(' ', spaces_replacement)
             for label in self.cell_indices.rows_labels[
                          # Reflects the row offset for export
                          self.export_offset[0]:self.shape[0]
                          ]
             ]
        if (y_helptext := self.cell_indices.rows_help_text) is not None:  # noqa E203
            y_helptext_descriptor = "row_description"
            # Reflects the row offset for export
            y_helptext = y_helptext[self.export_offset[0]:self.shape[0]]
        y_start_key = 'rows'

        # B) The x-axes represents the rows:
        if by_row:
            x_range = self.shape[0]
            x = [label.replace(' ', spaces_replacement)
                 for label in self.cell_indices.rows_labels[
                          # Reflects the row offset for export
                          self.export_offset[0]:self.shape[0]
                          ]
                 ]
            if (x_helptext := self.cell_indices.rows_help_text) is not None:  # noqa E203
                x_helptext_descriptor = "row_description"
                # Reflects the row offset for export
                x_helptext = x_helptext[self.export_offset[0]:self.shape[0]]
            x_start_key = 'rows'
            # The y-axes represents the columns
            y_range = self.shape[1]
            y = [label.replace(' ', spaces_replacement)
                 for label in self.cell_indices.columns_labels[
                          # Reflects the column offset for export
                          self.export_offset[1]:self.shape[1]
                          ]]
            if (y_helptext := self.cell_indices.columns_help_text) is not None:  # noqa E203
                y_helptext_descriptor = "column_description"
                # Reflects the column offset for export
                y_helptext = y_helptext[self.export_offset[1]:self.shape[1]]
            y_start_key = 'columns'

        # Export the spreadsheet to the dictionary (that can by JSON-ified)
        values = {x_start_key: {}}
        for idx_x in range(x_range):
            y_values = {y_start_key: {}}
            for idx_y in range(y_range):
                # Select the correct cell
                if by_row:
                    cell = self._get_cell_at(idx_x, idx_y)
                else:
                    cell = self._get_cell_at(idx_y, idx_x)
                # Skip if cell value is None if required:
                cell_value = cell.value
                if cell_value is None and skip_nan_cell:
                    continue
                # Replace the NaN value as required
                if cell_value is None:
                    cell_value = nan_replacement
                # Receive values from cell (either integer or building text)
                parsed_cell = cell.parse
                # Following dict is the dict that is exported as a cell
                pseudolang_and_val = {}
                cell_description: str = cell.description
                if cell_description is None:
                    # Replace cell description to NaN replacement
                    cell_description = nan_replacement
                    # If the cell description should be equal to some language
                    if descr_lang := use_language_for_description:  # noqa
                        cell_description = parsed_cell[descr_lang]
                # If it is an empty string, use None value instead
                if cell_description == "":
                    cell_description = nan_replacement
                # Add description in all languages wanted (by aliases)
                for i, language in enumerate(languages):
                    pseudolang_and_val[languages_used[i]] = \
                        parsed_cell[language]
                # Append the value:
                pseudolang_and_val['value'] = cell_value
                pseudolang_and_val['description'] = cell_description
                y_values[y_start_key][y[idx_y]] = pseudolang_and_val
                if y_helptext is not None:
                    # Add the help text for the row/column
                    if (ht := y_helptext[idx_y]) is None or len(  # noqa
                            y_helptext[idx_y]) == 0:
                        # If it is not set or is empty string, use NaN value
                        ht = nan_replacement
                    y_values[y_start_key][y[idx_y]][y_helptext_descriptor] = ht
            values[x_start_key][x[idx_x]] = y_values
            if x_helptext is not None:
                # Add the help text for the column/row
                if (ht := x_helptext[idx_x]) is None or len(  # noqa
                        x_helptext[idx_x]) == 0:
                    # If it is not set or is empty string, use NaN value
                    ht = nan_replacement
                values[x_start_key][x[idx_x]][x_helptext_descriptor] = ht

        # Create data parent
        data = {'data': values}

        # Add variables
        data['variables'] = self._get_variables().variables_dict
        # Add a row and column labels as arrays
        if by_row:
            data['row-labels'] = x
            data['column-labels'] = y
        else:
            data['row-labels'] = y
            data['column-labels'] = x

        # Create table parent
        table = {'table': data}

        # Append dictionary:
        for a_key, a_val in append_dict.items():
            table[a_key] = a_val

        return table

    def to_json(self,
                languages: List[str] = None,
                use_language_for_description: Optional[str] = None,
                /, *,  # noqa E999
                by_row: bool = True,
                languages_pseudonyms: List[str] = None,
                spaces_replacement: str = ' ',
                skip_nan_cell: bool = False,
                nan_replacement: object = None,
                append_dict: dict = {},
                generate_schema: bool = False) -> str:
        """Dumps the exported dictionary to the JSON object.

        Args:
            languages (List[str]): List of languages that should be exported.
                If it has value None, all the languages are exported.
            use_language_for_description (Optional[str]): If set-up (using
                the language name), description field is set to be either
                the description value (if defined) or the value of this
                language.
            by_row (bool): If True, rows are the first indices and columns
                are the second in the order. If False it is vice-versa.
            languages_pseudonyms (List[str]): Rename languages to the strings
                inside this list.
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            skip_nan_cell (bool): If True, None (NaN) values are skipped.
            nan_replacement (object): Replacement for the None (NaN) value
            append_dict (dict): Append this dictionary to output.
            generate_schema (bool): If true, returns the JSON schema.

        Returns:
            dict:
                Output dictionary with the logic:
                table->data->rows/columns->Row/Col label->
                columns/rows/help_text->Column/Row label->
                cell keys (value, description, language alias)->value
        """
        class _NumPyEncoder(json.JSONEncoder):
            """Encodes the NumPy variables to export"""
            def default(self, obj):
                if isinstance(obj, numpy.integer):
                    return int(obj)
                elif isinstance(obj, numpy.floating):
                    return float(obj)
                elif isinstance(obj, numpy.ndarray):
                    return obj.tolist()
                else:
                    return super(_NumPyEncoder, self).default(obj)
        # Return correctly encoded JSON
        return json.dumps(
            self.to_dictionary(languages, use_language_for_description,
                               by_row=by_row,
                               languages_pseudonyms=languages_pseudonyms,
                               spaces_replacement=spaces_replacement,
                               skip_nan_cell=skip_nan_cell,
                               nan_replacement=nan_replacement,
                               append_dict=append_dict,
                               generate_schema=generate_schema),
            cls=_NumPyEncoder
        )

    @staticmethod
    def generate_json_schema() -> dict:
        """Generate JSON schema.

        Returns:
            dict: JSON schema for the Portable Spreadsheet JSON table output.
        """
        false = False
        schema = {
            "$id": "http://portable-spreadsheet.com/spreadsheet.v2.schema.json",  # noqa
            # noqa
            "$schema": "http://json-schema.org/draft-07/schema#",
            "title": "Portable Spreadsheet JSON output schema",
            "description": "JSON schema of the Portable Spreadsheet output",

            "required": ["table"],

            "properties": {
                "table": {
                    "type": "object",
                    "required": ["variables", "row-labels", "column-labels",
                                 "data"],  # noqa
                    "properties": {
                        "row-labels": {
                            "minProperties": 0,
                            "propertyNames": {
                                "pattern": "^[0-9]*$"
                            },
                            "patternProperties": {
                                "^[0-9]*$": {
                                    "type": "string"
                                }
                            },
                            "additionalProperties": false
                        },
                        "column-labels": {
                            "minProperties": 0,
                            "propertyNames": {
                                "pattern": "^[0-9]*$"
                            },
                            "patternProperties": {
                                "^[0-9]*$": {
                                    "type": "string"
                                }
                            },
                            "additionalProperties": false
                        },
                        "variables": {
                            "minProperties": 0,
                            "propertyNames": {
                                "type": "string"
                            },
                            "patternProperties": {
                                "^.*$": {
                                    "type": "object",
                                    "properties": {
                                        "value": {
                                            "anyOf": [
                                                {"type": "string"},
                                                {"type": "number"},
                                                {"type": "null"}
                                            ]
                                        },
                                        "description": {
                                            "anyOf": [
                                                {"type": "string"},
                                                {"type": "number"},
                                                {"type": "null"}
                                            ]
                                        }
                                    },
                                    "required": ["value", "description"],
                                    "additionalProperties": false
                                }
                            },
                            "additionalProperties": false
                        },
                        "data": {
                            "type": "object",
                            "minProperties": 1,
                            "maxProperties": 2,
                            "additionalProperties": false,
                            "patternProperties": {
                                r"rows|columns": {
                                    "type": "object",
                                    "minProperties": 0,
                                    "propertyNames": {
                                        "type": "string"
                                    },
                                    "patternProperties": {
                                        "^.*$": {
                                            "type": "object",
                                            "patternProperties": {
                                                r"rows|columns": {
                                                    "type": "object",
                                                    "minProperties": 0,
                                                    "propertyNames": {
                                                        "type": "string"
                                                    },
                                                    "patternProperties": {
                                                        "^.*$": {
                                                            "type": "object",
                                                            "minProperties": 2,
                                                            "propertyNames": {
                                                                "type": "string" # noqa
                                                            },
                                                            "patternProperties": {  # noqa
                                                                "value": {
                                                                    "anyOf": [
                                                                        {"type": "string"},  # noqa
                                                                        {"type": "number"},  # noqa
                                                                        {"type": "null"}  # noqa
                                                                    ]
                                                                },
                                                                "description": {  # noqa
                                                                    "anyOf": [
                                                                        {"type": "string"},  # noqa
                                                                        {"type": "number"},  # noqa
                                                                        {"type": "null"}  # noqa
                                                                    ]
                                                                },
                                                                r"row_description|column_description": {  # noqa
                                                                    "anyOf": [
                                                                        {"type": "string"},  # noqa
                                                                        {"type": "number"},  # noqa
                                                                        {"type": "null"}  # noqa
                                                                    ]
                                                                },
                                                                "^.*$": {
                                                                    "anyOf": [
                                                                        {"type": "string"},  # noqa
                                                                        {"type": "number"},  # noqa
                                                                        {"type": "null"}  # noqa
                                                                    ]
                                                                },
                                                            },
                                                            "required": ['value', 'description'],  # noqa
                                                            "additionalProperties": false  # noqa
                                                        }
                                                    }
                                                },
                                                r"row_description|column_description": {  # noqa
                                                    "anyOf": [
                                                            {"type": "string"},
                                                            {"type": "number"},
                                                            {"type": "null"}
                                                        ]
                                                    }
                                            },
                                            "minProperties": 1,
                                            "maxProperties": 2,
                                            "additionalProperties": false
                                        }
                                    },
                                    "additionalProperties": false
                                }
                            }
                        }
                    },
                    "additionalProperties": false
                }
            },
            "additionalProperties": {
                "type": "object"
            }
        }

        return schema

    def to_string_of_values(self) -> str:
        """Export values inside table to the Python array definition string.

        Returns:
            str: Python list definition string.
        """
        # Log warning if needed
        self.log_export_subset_warning_if_needed()

        export = "["
        for row_idx in range(self.shape[0]):
            export += "["
            for col_idx in range(self.shape[1]):
                export += str(self._get_cell_at(row_idx, col_idx).value)
                if col_idx < self.shape[1] - 1:
                    export += ', '
            export += "]"
            if row_idx < self.shape[0] - 1:
                export += ",\n"
        return export + "]"

    def to_2d_list(self, *,
                   language: Optional[str] = None,
                   top_right_corner_text: str = "Sheet",
                   skip_labels: bool = False,
                   na_rep: Optional[object] = None,
                   spaces_replacement: str = ' ',) -> List[List[object]]:
        """Export values 2 dimensional Python array.

        Args:
            language (Optional[str]): If set-up, export the word in this
                language in each cell instead of values.
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            top_right_corner_text (str): Text in the top right corner.
            na_rep (str): Replacement for the missing data (cells with value
                equals to None).
            skip_labels (bool): If true, first row and column with labels is
                skipped

        Returns:
            List[List[object]]: Python array.
        """
        # Log warning if needed
        self.log_export_subset_warning_if_needed()

        export: list = []
        for row_idx in range(-1, self.shape[0]):
            row: list = []
            if row_idx == -1:
                if skip_labels:
                    continue
                row.append(top_right_corner_text)
                # Insert labels of columns:
                for col_i in range(self.shape[1]):
                    col = self.cell_indices.columns_labels[
                        col_i + self.export_offset[1]
                    ]
                    value = col.replace(' ', spaces_replacement)
                    row.append(value)
            else:
                if not skip_labels:
                    # Insert labels of rows
                    row.append(self.cell_indices.rows_labels[
                                  row_idx + self.export_offset[0]
                              ].replace(' ', spaces_replacement))
                for col_idx in range(self.shape[1]):
                    # Append actual values:
                    cell_at_position = self._get_cell_at(row_idx, col_idx)

                    if language is not None:
                        # Get the word of cell on current position
                        value_to_write = cell_at_position.parse[language]
                    else:
                        # Get the value of cell on current position
                        value_to_write = cell_at_position.value
                    if value_to_write is None:
                        # Replacement for None
                        value_to_write = na_rep
                    row.append(value_to_write)

            export.append(row)
        return export

    def to_csv(self, *,
               language: Optional[str] = None,
               top_right_corner_text: str = "Sheet",
               skip_labels: bool = False,
               na_rep: Optional[object] = '',
               spaces_replacement: str = ' ',

               sep: str = ',',
               line_terminator: str = '\n',
               ) -> str:
        """Export values to the string in the CSV logic

        Args:
            language (Optional[str]): If set-up, export the word in this
                language in each cell instead of values.
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            top_right_corner_text (str): Text in the top right corner.
            sep (str): Separator of values in a row.
            line_terminator (str): Ending sequence (character) of a row.
            na_rep (str): Replacement for the missing data.
            skip_labels (bool): If true, first row and column with labels is
                skipped

        Returns:
            str: CSV of the values
        """
        sheet_as_array = self.to_2d_list(
            top_right_corner_text=top_right_corner_text, na_rep=na_rep,
            skip_labels=skip_labels, spaces_replacement=spaces_replacement,
            language=language
        )
        export = ""
        for row_idx in range(len(sheet_as_array)):
            for col_idx in range(len(sheet_as_array[row_idx])):
                export += str(sheet_as_array[row_idx][col_idx])
                if col_idx < (len(sheet_as_array[row_idx]) - 1):
                    export += sep
            if row_idx < (len(sheet_as_array) - 1):
                export += line_terminator
        return export

    def to_markdown(self, *,
                    language: Optional[str] = None,
                    top_right_corner_text: str = "Sheet",
                    skip_labels: bool = False,
                    na_rep: Optional[object] = '',
                    spaces_replacement: str = ' '
                    ):
        """Export values to the string in the Markdown (MD) file logic

        Args:
            language (Optional[str]): If set-up, export the word in this
                language in each cell instead of values.
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            top_right_corner_text (str): Text in the top right corner.
            na_rep (str): Replacement for the missing data.
            skip_labels (bool): If true, first row and column with labels is
                skipped

        Returns:
            str: Markdown (MD) compatible table of the values
        """
        sheet_as_array = self.to_2d_list(
            top_right_corner_text=top_right_corner_text, na_rep=na_rep,
            skip_labels=skip_labels, spaces_replacement=spaces_replacement,
            language=language
        )
        export = ""
        for row_idx in range(len(sheet_as_array)):
            # Add values:
            export += "| "
            for col_idx in range(len(sheet_as_array[row_idx])):
                if (row_idx == 0 or col_idx == 0) and not skip_labels:
                    export += '*'
                export += str(sheet_as_array[row_idx][col_idx])
                if (row_idx == 0 or col_idx == 0) and not skip_labels:
                    export += '*'
                if col_idx < (len(sheet_as_array[row_idx]) - 1):
                    export += " | "
                else:
                    export += " |"
            export += "\n"
            # Add |---| sequence after the first line
            if row_idx == 0 and not skip_labels:
                for col_idx in range(len(sheet_as_array[row_idx])):
                    export += "|----"
                    if col_idx == (len(sheet_as_array[row_idx]) - 1):
                        export += '|'
                export += "\n"
        # Add first two lines if labels are skipped
        if skip_labels:
            # Add empty || separators for missing labels
            first_line = "|"
            for col_idx in range(len(sheet_as_array[0])):
                first_line += "|"
            first_line += "|\n"
            # Add |---| sequence after the first line
            for col_idx in range(len(sheet_as_array[0])):
                first_line += "|----"
                if col_idx == (len(sheet_as_array[0]) - 1):
                    first_line += '|'
            export = first_line + "\n" + export

        return export

    def to_numpy(self) -> numpy.ndarray:
        """Exports the values to the numpy.ndarray.

        Returns:
            numpy.ndarray: 2 dimensions array with values
        """
        # Log warning if needed
        self.log_export_subset_warning_if_needed()

        results = numpy.zeros(self.shape)
        # Variable for indicating that logging is needed (for logging that
        # replacement of some value for NaN is done):
        contains_nonumeric_values = False
        for row_idx in range(self.shape[0]):
            for col_idx in range(self.shape[1]):
                if (value := self._get_cell_at(row_idx, col_idx).value) is not None:  # noqa E999
                    if isinstance(value, Number):
                        results[row_idx, col_idx] = value
                    else:
                        results[row_idx, col_idx] = numpy.nan
                        # For logging that replacement is done
                        contains_nonumeric_values = True
                else:
                    results[row_idx, col_idx] = numpy.nan
        # Log warning if needed
        if contains_nonumeric_values:
            self.warning_logger(
                "Some values in the sheet are not numbers, the "
                "nan value is set instead."
            )
        return results

    def to_html_table(self, *,
                      spaces_replacement: str = ' ',
                      top_right_corner_text: str = "Sheet",
                      na_rep: str = '',
                      language_for_description: str = None,
                      skip_labels: bool = False) -> str:
        """Export values to the string in the HTML table logic

        Args:
            spaces_replacement (str): All the spaces in the rows and columns
                descriptions (labels) are replaced with this string.
            top_right_corner_text (str): Text in the top right corner.
            na_rep (str): Replacement for the missing data.
            language_for_description (str): If not None, the description
                of each computational cell is inserted as word of this
                language (if the property description is not set).
            skip_labels (bool): If true, first row and column with labels is
                skipped

        Returns:
            str: HTML table definition
        """
        # Log warning if needed
        self.log_export_subset_warning_if_needed()

        export = "<table>"
        for row_idx in range(-1, self.shape[0]):
            if skip_labels and row_idx == -1:
                continue
            export += "<tr>"
            if row_idx == -1:
                export += "<th>"
                export += top_right_corner_text
                export += "</th>"
                # Insert labels of columns:
                for col_i in range(self.shape[1]):
                    export += "<th>"
                    col = self.cell_indices.columns_labels[
                        col_i + self.export_offset[1]
                        ]
                    if (help_text :=  # noqa 203
                            self.cell_indices.columns_help_text) is not None:
                        title_attr = ' title="{}"'.format(
                            help_text[col_i + self.export_offset[1]]
                        )
                    else:
                        title_attr = ""
                    export += f'<a href="javascript:;" {title_attr}>'
                    export += col.replace(' ', spaces_replacement)
                    export += "</a>"
                    export += "</th>"
            else:
                if not skip_labels:
                    # Insert labels of rows
                    if (help_text :=  # noqa: 203
                            self.cell_indices.rows_help_text) is not None:
                        title_attr = ' title="{}"'.format(
                            help_text[row_idx + self.export_offset[1]]
                        )
                    else:
                        title_attr = ""
                    export += "<td>"
                    export += f'<a href="javascript:;" {title_attr}>'
                    export += self.cell_indices.rows_labels[
                                  row_idx + self.export_offset[0]
                                  ].replace(' ', spaces_replacement)
                    export += "</a>"
                    export += "</td>"

                # Insert actual values in the spreadsheet
                for col_idx in range(self.shape[1]):
                    title_attr = ""
                    cell_at_pos = self._get_cell_at(row_idx, col_idx)
                    if (description := # noqa 203
                            cell_at_pos.description) is not None:
                        title_attr = ' title="{}"'.format(description)
                    elif language_for_description is not None:
                        if cell_at_pos.cell_type == CellType.computational:
                            title = cell_at_pos.constructing_words.words[
                                language_for_description
                            ]
                            title_attr = f' title="{title}"'
                    export += "<td>"
                    export += f'<a href="javascript:;" {title_attr}>'
                    value = cell_at_pos.value
                    if value is None:
                        value = na_rep
                    export += str(value)
                    export += "</a>"
                    export += "</td>"
            export += '</tr>'
        export += '</table>'
        return export

    @property
    def columns(self) -> List[str]:
        """Return the labels of columns.

        Returns:
            List[str]: List of column labels
        """
        return self.cell_indices.columns_labels[
            self.export_offset[1]:(self.export_offset[1] + self.shape[1])
        ]

    @property
    def index(self) -> List[str]:
        """Return the labels of rows.

        Returns:
            List[str]: List of row labels.
        """
        return self.cell_indices.rows_labels[
            self.export_offset[0]:(self.export_offset[0] + self.shape[0])
        ]



from typing import Set

from .grammars import GRAMMAR_PATTERN, GRAMMARS


class GrammarUtils(object):
    """Utils for validating and adding user defined grammars to the system.
    """

    @staticmethod
    def _get_value_at_path(grammar: dict, path: list):
        """Get the value on the path in the dictionary of logic defined in the
            path ['a', 'b'].

        Args:
            grammar (dict): Grammar dictionary.
            path (list): Path inside the grammar (1D array)

        Returns:
            object: value inside
        """
        if len(path) == 0:
            return grammar
        else:
            start, *rest = path
            return GrammarUtils._get_value_at_path(grammar[start], rest)

    @staticmethod
    def _check_dictionary(grammar: dict, key: list = []):
        """Walk across all values in the grammar and check if types of all
            words definitions matches.

        Args:
            grammar (dict): Probed grammar.
            key (list): The path to the probed entity.
        """
        # Possibly dictionary, possibly value:
        _pos_dict = GrammarUtils._get_value_at_path(grammar, key)
        if isinstance(_pos_dict, dict):
            _keys = _pos_dict.keys()
            for _key in _keys:
                GrammarUtils._check_dictionary(grammar, key + [_key])
        else:
            # Expected type
            ex_type = GrammarUtils._get_value_at_path(GRAMMAR_PATTERN, key)
            if isinstance(_pos_dict, ex_type):
                pass
            else:
                raise ValueError(f"Input grammar is not valid at path {key}!")

    @staticmethod
    def validate_grammar(grammar: dict, raise_exception: bool = False) -> bool:
        """Validate the input grammar.

        Args:
            grammar (dict): Grammar definition.
            raise_exception (bool): If True, error is raised when occurs.

        Returns:
            bool: True, if the grammar is valid, False otherwise.

        Raises:
            ValueError: If grammar is invalid and argument raise_exception.
        """
        try:
            GrammarUtils._check_dictionary(grammar)
        except ValueError as ve:
            if raise_exception:
                raise ve
            else:
                return False
        except KeyError as ke:
            if raise_exception:
                raise ValueError("Invalid keys in input!") from ke
            else:
                return False
        return True

    @staticmethod
    def add_grammar(grammar_definition: dict, language_name: str) -> None:
        """Add the grammar to the system.

        Args:
            grammar_definition (dict): Definition of the grammar.
            language_name (str): Name of the language (like 'excel', 'python')
        """
        # Check the grammar and raise exception if it is invalid
        GrammarUtils.validate_grammar(grammar_definition, True)
        # Add the grammar
        if language_name not in GRAMMARS.keys():
            GRAMMARS[language_name] = grammar_definition
        else:
            raise ValueError(f"Language {language_name} is already in the "
                             "system.")

    @staticmethod
    def remove_grammar(language_name: str) -> None:
        """Remove the grammar from the system.

        Args:
            language_name (str): Name of the language (like 'excel', 'python')
        """
        # Add the grammar
        if language_name in GRAMMARS.keys():
            del GRAMMARS[language_name]
        else:
            raise ValueError(f"Language {language_name} is not in the "
                             "system.")

    @staticmethod
    def get_languages() -> Set[str]:
        """Returns the set of included languages.

        Returns:
            Set[str]: Set of all languages included in the system.
        """
        return set([str(lang) for lang in GRAMMARS.keys()])

    @staticmethod
    def check_system_consistency() -> bool:
        """Check if the system contains all neccessary grammars.

        Returns:
            bool: True if system is consistent, False otherwise
        """
        languages_in_system = GrammarUtils.get_languages()
        if len({'excel', 'python_numpy'} & languages_in_system) != 2:
            return False
        validation = GrammarUtils.validate_grammar(GRAMMARS['excel']) \
            and GrammarUtils.validate_grammar(GRAMMARS['python_numpy'])
        return validation



import unittest

from typing import Callable, Collection

import numpy_financial as npf
import numpy as np

from portable_spreadsheet.cell import Cell
from portable_spreadsheet.cell_indices import CellIndices
from portable_spreadsheet.cell_type import CellType


class TestCellBasicFunctionality(unittest.TestCase):
    """Integration test for basic Cell functionality and WordConstructor
    functionality.

    Notice:
        if variable start with prefix 'a_' it means that cell is anchored
        (located in grid), if with prefix 'u_' it means that it is NOT
        anchored, with prefix 't' it does not matter (just a testing variable).
    """
    def setUp(self) -> None:
        self.cell_indices: CellIndices = CellIndices(5, 7)

    def test_initialisation(self):
        """Test the constructor of class Cell"""
        with self.assertRaises(ValueError):
            Cell(3, None, cell_indices=self.cell_indices)

        with self.assertRaises(AttributeError):
            Cell(3, 3, cell_indices=None)

        # Create some reasonable cell
        t_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        self.assertEqual(t_cell.row, 3)
        self.assertEqual(t_cell.column, 4)
        self.assertEqual(t_cell.value, 7)
        self.assertEqual(t_cell.cell_type, CellType.value_only)
        self.assertFalse(t_cell.is_variable)
        self.assertIsNone(t_cell.variable_name)
        self.assertEqual(len(t_cell._excel_format), 0)
        self.assertIsNone(t_cell._description)
        # Check the initialisation words
        words = t_cell.constructing_words
        self.assertEqual(words.words['excel'], '7')
        self.assertEqual(words.words['python_numpy'], '7')
        self.assertTrue(t_cell.anchored)

    def test_word_property(self):
        """Test the property 'word'"""
        # Anchored cell
        a_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        # There is an offset 1 row/column for labels!
        self.assertEqual(a_cell.word.words['excel'], 'F5')
        self.assertEqual(a_cell.word.words['python_numpy'], 'values[3,4]')
        self.assertTrue(a_cell.anchored)

        # Un-anchored cell
        u_cell = Cell(value=7, cell_indices=self.cell_indices)
        self.assertEqual(u_cell.word.words['excel'], '7')
        self.assertEqual(u_cell.word.words['python_numpy'], '7')
        self.assertFalse(u_cell.anchored)

    def test_coordinates_property(self):
        """Test coordinates"""
        # Anchored cell
        a_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        self.assertTupleEqual(a_cell.coordinates, (3, 4))

        # Un-anchored cell
        u_cell = Cell(value=7, cell_indices=self.cell_indices)
        self.assertTupleEqual(u_cell.coordinates, (None, None))
        # Test setter
        u_cell.coordinates = (3, 2)
        self.assertTupleEqual(u_cell.coordinates, (3, 2))

    def test_value_property(self):
        """Test the value property"""
        t_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        self.assertEqual(t_cell.value, 7)

    def test_parse_property(self):
        """Test the parsing (basically tests the method parse from the
            WordConstructor class
        """
        # Anchored & Value cells
        a_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        a_parsed = a_cell.parse
        self.assertEqual(a_parsed['excel'], '7')
        self.assertEqual(a_parsed['python_numpy'], '7')
        # empty cell
        e_cell = Cell(3, 4, None, cell_indices=self.cell_indices)
        e_parsed = e_cell.parse
        self.assertEqual(e_parsed['excel'], '')
        self.assertEqual(e_parsed['python_numpy'], '')
        # Computational cells
        # have to do some simple computation
        operand_1 = Cell(3, 4, 7, cell_indices=self.cell_indices)
        operand_2 = Cell(2, 4, 6, cell_indices=self.cell_indices)
        sum_1_2 = operand_1 + operand_2
        s_parsed = sum_1_2.parse
        self.assertEqual(s_parsed['excel'], '=F5+F4')
        self.assertEqual(s_parsed['python_numpy'], 'values[3,4]+values[2,4]')

    def test_excel_format(self):
        """Test if excel format getter/setter works"""
        t_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        with self.assertRaises(ValueError):
            t_cell.excel_format = 7
        t_cell.excel_format = {'bold': True}
        self.assertDictEqual(t_cell.excel_format, {'bold': True})
        self.assertDictEqual(t_cell._excel_format, {'bold': True})

    def test_description(self):
        """Test if the description works"""
        t_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        with self.assertRaises(ValueError):
            t_cell.description = 7
        t_cell.description = "Super description"
        self.assertEqual(t_cell.description, "Super description")
        self.assertEqual(t_cell._description, "Super description")

    def test_constructing_words(self):
        """Test the constructing words property."""
        t_cell = Cell(3, 4, 7, cell_indices=self.cell_indices)
        self.assertEqual(t_cell.constructing_words, t_cell._constructing_words)

    def test_conditional(self):
        """Regression test for the conditional."""
        a_cell_cond_1 = Cell(1, 3, 8, cell_indices=self.cell_indices)
        u_cell_cond_2 = Cell(value=7, cell_indices=self.cell_indices)
        a_cell_conseq = Cell(2, 3, 11, cell_indices=self.cell_indices)
        u_cell_altern = Cell(value=5, cell_indices=self.cell_indices)

        u_res = Cell.conditional(a_cell_cond_1 == u_cell_cond_2,  # <- IF
                                 a_cell_conseq,  # <- THEN
                                 u_cell_altern)  # <- ELSE
        # Test generated words
        res_word = u_res.parse
        word_expected = {
            'python_numpy': '((values[2,3]) if (values[1,3]==7) else (5))',
            'excel': '=IF(E3=7,E4,5)'
        }
        self.assertDictEqual(word_expected, res_word)
        # Test generated values
        value_computed = u_res.value
        value_expected = u_cell_altern.value
        self.assertEqual(value_expected, value_computed)

    def test_offset(self):
        """Regression test for the offset."""
        a_cell_ref = Cell(1, 3, 8, cell_indices=self.cell_indices)
        u_cell_row = Cell(value=7, cell_indices=self.cell_indices)
        a_cell_col = Cell(2, 3, 11, cell_indices=self.cell_indices)
        u_cell_tar = Cell(5, 5, 5, cell_indices=self.cell_indices)

        u_res = Cell.offset(a_cell_ref, u_cell_row, a_cell_col,
                            target=u_cell_tar)
        # Test generated words
        res_word = u_res.parse
        word_expected = {'excel': '=OFFSET(E3,7,E4)',
                         'python_numpy': 'values[1+7,3+values[2,3]]'}
        self.assertDictEqual(word_expected, res_word)
        # Test generated values
        value_computed = u_res.value
        value_expected = u_cell_tar.value
        self.assertEqual(value_expected, value_computed)


class TestCellBinaryOperations(unittest.TestCase):
    """Test binary operations.

    Notice:
        if variable start with prefix 'a_' it means that cell is anchored
        (located in grid), if with prefix 'u_' it means that it is NOT
        anchored, with prefix 't' it does not matter (just a testing variable).
    """
    def setUp(self) -> None:
        self.cell_indices: CellIndices = CellIndices(5, 7)
        self.a_operand_1 = Cell(3, 4, 7, cell_indices=self.cell_indices)
        self.a_operand_2 = Cell(2, 4, 4, cell_indices=self.cell_indices)
        self.coord_operand_1_python = "values[3,4]"
        self.coord_operand_2_python = "values[2,4]"
        self.coord_operand_1_excel = "F5"
        self.coord_operand_2_excel = "F4"

        self.u_operand_1 = Cell(value=7, cell_indices=self.cell_indices)
        self.u_operand_2 = Cell(value=8, cell_indices=self.cell_indices)
        self.value_u_1 = "7"
        self.value_u_2 = "8"

    def _check_binary_operation(
            self,
            operation_method: Callable[[Cell, Cell], Cell],
            real_operation_fn: Callable[[float, float], float],
            excel_separator: str,
            python_numpy_separator: str, *,
            excel_prefix: str = "",
            excel_suffix: str = "",
            excel_reference_prefix: str = "=",
            python_reference_prefix: str = "",
    ) -> None:
        """Run the binary operation, compare numeric result, compare words.

        Args:
            operation_method (Callable[[Cell, Cell], Cell]): Pointer to the
                method inside the Cell class.
            real_operation_fn (Callable[[float, float], float]): Pointer to the
                Python method that compute the same.
            excel_separator (str): Separator of the operation in
                Excel language.
            python_numpy_separator (str): Separator of the operation in
                Python_NumPY language.
            excel_prefix (str): Prefix for the operation in Excel language.
            excel_suffix (str): Suffix for the operation in Excel language.
            excel_reference_prefix (str): Prefix of the word that reference
                to some computation in Excel.
            python_reference_prefix (str): Prefix of the word that reference
                to some computation in Python.
        """
        # A) Anchored
        a_res_cell = operation_method(self.a_operand_1, self.a_operand_2)
        # Compare words
        a_res_parsed = a_res_cell.parse
        self.assertEqual(a_res_parsed['excel'],
                         (excel_reference_prefix + excel_prefix +
                          self.coord_operand_1_excel + excel_separator +
                          self.coord_operand_2_excel + excel_suffix)
                         )
        self.assertEqual(a_res_parsed['python_numpy'],
                         (python_reference_prefix +
                          self.coord_operand_1_python +
                          python_numpy_separator +
                          self.coord_operand_2_python)
                         )
        # Compare results of anchored
        self.assertAlmostEqual(a_res_cell.value,
                               real_operation_fn(self.a_operand_1.value,
                                                 self.a_operand_2.value))
        # B) Un-anchored
        u_res_cell = operation_method(self.u_operand_1, self.u_operand_2)
        # Compare words
        u_res_parsed = u_res_cell.parse
        self.assertEqual(u_res_parsed['excel'],
                         (excel_reference_prefix + excel_prefix +
                          self.value_u_1 +
                          excel_separator +
                          self.value_u_2 + excel_suffix)
                         )
        self.assertEqual(u_res_parsed['python_numpy'],
                         (python_reference_prefix +
                          self.value_u_1 +
                          python_numpy_separator +
                          self.value_u_2)
                         )
        # Compare results of un-anchored
        self.assertAlmostEqual(u_res_cell.value,
                               real_operation_fn(self.u_operand_1.value,
                                                 self.u_operand_2.value))

    def test_add(self):
        """Test adding"""
        # Method test
        self._check_binary_operation(Cell.add,
                                     lambda x, y: x + y,
                                     "+",
                                     "+")
        # Operator test
        self._check_binary_operation(Cell.__add__,
                                     lambda x, y: x + y,
                                     "+",
                                     "+")

    def test_subtract(self):
        """Test subtracting"""
        # Method test
        self._check_binary_operation(Cell.subtract,
                                     lambda x, y: x - y,
                                     "-",
                                     "-")
        # Operator test
        self._check_binary_operation(Cell.__sub__,
                                     lambda x, y: x - y,
                                     "-",
                                     "-")

    def test_multiply(self):
        """Test multiplying"""
        # Method test
        self._check_binary_operation(Cell.multiply,
                                     lambda x, y: x * y,
                                     "*",
                                     "*")
        # Operator test
        self._check_binary_operation(Cell.__mul__,
                                     lambda x, y: x * y,
                                     "*",
                                     "*")

    def test_divide(self):
        """Test dividing"""
        # Method test
        self._check_binary_operation(Cell.divide,
                                     lambda x, y: x / y,
                                     "/",
                                     "/")
        # Operator test
        self._check_binary_operation(Cell.__truediv__,
                                     lambda x, y: x / y,
                                     "/",
                                     "/")

    def test_modulo(self):
        """Test modulo"""
        # Method test
        self._check_binary_operation(Cell.modulo,
                                     lambda x, y: x % y,
                                     ",",
                                     "%",
                                     excel_prefix="MOD(",
                                     excel_suffix=")")
        # Operator test
        self._check_binary_operation(Cell.__mod__,
                                     lambda x, y: x % y,
                                     ",",
                                     "%",
                                     excel_prefix="MOD(",
                                     excel_suffix=")")

    def test_power(self):
        """Test power"""
        # Method test
        self._check_binary_operation(Cell.power,
                                     lambda x, y: x ** y,
                                     "^",
                                     "**")
        # Operator test
        self._check_binary_operation(Cell.__pow__,
                                     lambda x, y: x ** y,
                                     "^",
                                     "**")

    def test_equalTo(self):
        """Test equal to"""
        # Method test
        self._check_binary_operation(Cell.equalTo,
                                     lambda x, y: x == y,
                                     "=",
                                     "==")
        # Operator test
        self._check_binary_operation(Cell.__eq__,
                                     lambda x, y: x == y,
                                     "=",
                                     "==")

    def test_notEqualTo(self):
        """Test not equal to"""
        # Method test
        self._check_binary_operation(Cell.notEqualTo,
                                     lambda x, y: x != y,
                                     "<>",
                                     "!=")
        # Operator test
        self._check_binary_operation(Cell.__ne__,
                                     lambda x, y: x != y,
                                     "<>",
                                     "!=")

    def test_greaterThan(self):
        """Test greater than"""
        # Method test
        self._check_binary_operation(Cell.greaterThan,
                                     lambda x, y: x > y,
                                     ">",
                                     ">")
        # Operator test
        self._check_binary_operation(Cell.__gt__,
                                     lambda x, y: x > y,
                                     ">",
                                     ">")

    def test_greaterThanOrEqualTo(self):
        """Test greater than or equal to"""
        # Method test
        self._check_binary_operation(Cell.greaterThanOrEqualTo,
                                     lambda x, y: x >= y,
                                     ">=",
                                     ">=")
        # Operator test
        self._check_binary_operation(Cell.__ge__,
                                     lambda x, y: x >= y,
                                     ">=",
                                     ">=")

    def test_lessThan(self):
        """Test less than"""
        # Method test
        self._check_binary_operation(Cell.lessThan,
                                     lambda x, y: x < y,
                                     "<",
                                     "<")
        # Operator test
        self._check_binary_operation(Cell.__lt__,
                                     lambda x, y: x < y,
                                     "<",
                                     "<")

    def test_lessThanOrEqualTo(self):
        """Test less than or equal to"""
        # Method test
        self._check_binary_operation(Cell.lessThanOrEqualTo,
                                     lambda x, y: x <= y,
                                     "<=",
                                     "<=")
        # Operator test
        self._check_binary_operation(Cell.__le__,
                                     lambda x, y: x <= y,
                                     "<=",
                                     "<=")

    def test_LogicalConjunction(self):
        """Test logical conjunction"""
        # Method test
        self._check_binary_operation(Cell.logicalConjunction,
                                     lambda x, y: x and y,
                                     ", ",
                                     " and ",
                                     excel_prefix="AND(",
                                     excel_suffix=")")
        # Operator test
        self._check_binary_operation(Cell.__and__,
                                     lambda x, y: x and y,
                                     ", ",
                                     " and ",
                                     excel_prefix="AND(",
                                     excel_suffix=")")

    def test_LogicalDisjunction(self):
        """Test logical disjunction"""
        # Method test
        self._check_binary_operation(Cell.logicalDisjunction,
                                     lambda x, y: x or y,
                                     ", ",
                                     " or ",
                                     excel_prefix="OR(",
                                     excel_suffix=")")
        # Operator test
        self._check_binary_operation(Cell.__or__,
                                     lambda x, y: x or y,
                                     ", ",
                                     " or ",
                                     excel_prefix="OR(",
                                     excel_suffix=")")

    def test_chain_of_operation(self):
        """Test multiple operations in a sequence"""
        result = self.a_operand_1 + self.a_operand_2 * self.u_operand_1
        self.assertDictEqual({'python_numpy': 'values[3,4]+values[2,4]*7',
                              'excel': '=F5+F4*7'},
                             result.parse)
        self.assertAlmostEqual(result.value, 35)

    def test_raw_statement(self):
        """Test raw statement"""
        result = Cell.raw(self.a_operand_1, {
            'python_numpy': "Hello from Python",
            'excel': "Excel welcomes"
        })
        self.assertDictEqual({
            'python_numpy': "Hello from Python",
            'excel': "=Excel welcomes"  # Always computational
        }, result.parse)
        self.assertAlmostEqual(self.a_operand_1.value, 7)

    def test_concatenate(self):
        """Test string concatenation."""
        # Define grammar rules
        excel_reference_prefix = "="
        excel_prefix = "CONCATENATE("
        excel_separator = ","
        excel_suffix = ")"
        python_reference_prefix = ""
        python_numpy_separator = "+"
        # A) Anchored
        a_res_cell = self.a_operand_1.concatenate(self.a_operand_2)
        # Compare words
        a_res_parsed = a_res_cell.parse
        self.assertEqual((excel_reference_prefix + excel_prefix +
                          self.coord_operand_1_excel + excel_separator +
                          self.coord_operand_2_excel + excel_suffix),
                         a_res_parsed['excel'])
        self.assertEqual((python_reference_prefix +
                          self.coord_operand_1_python +
                          python_numpy_separator +
                          self.coord_operand_2_python),
                         a_res_parsed['python_numpy']
                         )
        # Compare results of anchored
        self.assertAlmostEqual(a_res_cell.value,
                               str(self.a_operand_1.value) +
                               str(self.a_operand_2.value))

        # B) Un-anchored, numerical
        u_res_cell = self.u_operand_1.concatenate(self.u_operand_2)
        # Compare words
        u_res_parsed = u_res_cell.parse
        self.assertEqual(u_res_parsed['excel'],
                         (excel_reference_prefix +
                          excel_prefix +
                          self.value_u_1 +
                          excel_separator +
                          self.value_u_2 +
                          excel_suffix)
                         )
        self.assertEqual(u_res_parsed['python_numpy'],
                         ('"' + python_reference_prefix +
                          self.value_u_1 + '"' +
                          python_numpy_separator +
                          '"' + self.value_u_2 + '"')
                         )
        # Compare results of un-anchored
        self.assertAlmostEqual(u_res_cell.value,
                               str(self.u_operand_1.value) +
                               str(self.u_operand_2.value))

        # C) Un-anchored, strings
        str_value = "6ppW1lPT"
        u_operand_1 = Cell(value=str_value, cell_indices=self.cell_indices)
        u_res_cell = u_operand_1.concatenate(self.u_operand_2)
        # Compare words
        u_res_parsed = u_res_cell.parse
        self.assertEqual(u_res_parsed['excel'],
                         (excel_reference_prefix +
                          excel_prefix +
                          '"' + str_value + '"' +
                          excel_separator +
                          self.value_u_2 +
                          excel_suffix)
                         )
        self.assertEqual(u_res_parsed['python_numpy'],
                         (python_reference_prefix +
                          str_value +
                          python_numpy_separator +
                          '"' + self.value_u_2 + '"')
                         )
        # Compare results of un-anchored
        self.assertAlmostEqual(u_res_cell.value,
                               u_operand_1.value +
                               str(self.u_operand_2.value))


class TestCellAggregationFunctionality(unittest.TestCase):
    """Test aggregation functions.

    Notice:
        if variable start with prefix 'a_' it means that cell is anchored
        (located in grid), if with prefix 'u_' it means that it is NOT
        anchored, with prefix 't' it does not matter (just a testing variable).
    """
    def setUp(self) -> None:
        self.cell_indices: CellIndices = CellIndices(5, 7)
        self.a_cell_start = Cell(1, 2, 7, cell_indices=self.cell_indices)
        self.a_cell_end = Cell(3, 5, 4, cell_indices=self.cell_indices)
        # Words for slices (aggregation operation):
        self.slice_python = "values[1:4,2:6]"
        self.slice_excel = "D3:G5"  # shifted because of offset to labels
        # Create the set (collection of values)
        self.slice_cardinality = 12
        self.cell_set = [Cell(i, j, np.random.random() * 10,
                              cell_indices=self.cell_indices)
                         for i in range(1, 3 + 1)
                         for j in range(2, 5 + 1)
                         ]
        self.cell_set[0]._value = -100
        self.cell_values = [self.cell_set[i].value
                            for i in range(self.slice_cardinality)]

    def _check_aggregate_function(
            self,
            operation_method: Callable[[Cell, Cell, Collection[Cell]], Cell],
            real_operation_fn: Callable[[float, float], float],
            excel_prefix: str,
            excel_suffix: str,
            python_prefix: str,
            python_suffix: str, *,
            excel_reference_prefix: str = "=",
            python_reference_prefix: str = "",
    ) -> None:
        """Run the aggregate function, compare numeric result, compare words.

        Args:
            operation_method (Callable[[Cell, Cell], Cell], Collection[Cell]):
                Pointer to the method inside the Cell class.
            real_operation_fn (Callable[[float, float], float]): Pointer to the
                Python method that compute the same.
            excel_prefix (str): Prefix for the operation in Excel language.
            excel_suffix (str): Suffix for the operation in Excel language.
            python_prefix (str): Prefix for the operation in Python_NumPy
                language.
            python_suffix (str): Suffix for the operation in Python_NumPy
                language.
            excel_reference_prefix (str): Prefix of the word that reference
                to some computation in Excel.
            python_reference_prefix (str): Prefix of the word that reference
                to some computation in Python.
        """
        u_result = operation_method(self.a_cell_start, self.a_cell_end,
                                    self.cell_set)
        # Check the values
        u_value_computed = u_result.value
        value_expected = real_operation_fn(self.cell_values)
        self.assertAlmostEqual(u_value_computed, value_expected)

        # Compare words
        u_res_parsed = u_result.parse
        self.assertEqual(u_res_parsed['excel'],
                         (excel_reference_prefix + excel_prefix +
                          self.slice_excel + excel_suffix)
                         )
        self.assertEqual(u_res_parsed['python_numpy'],
                         (python_reference_prefix + python_prefix +
                          self.slice_python + python_suffix)
                         )

    def test_sum(self):
        """Test the sum"""
        self._check_aggregate_function(Cell.sum, np.sum, "SUM(", ")",
                                       "np.sum(", ")")

    def test_product(self):
        """Test the product"""
        self._check_aggregate_function(Cell.product, np.prod, "PRODUCT(", ")",
                                       "np.prod(", ")")

    def test_mean(self):
        """Test the mean"""
        self._check_aggregate_function(Cell.mean, np.mean, "AVERAGE(", ")",
                                       "np.mean(", ")")

    def test_min(self):
        """Test the min"""
        self._check_aggregate_function(Cell.min, np.min, "MIN(", ")",
                                       "np.min(", ")")

    def test_max(self):
        """Test the max"""
        self._check_aggregate_function(Cell.max, np.max, "MAX(", ")",
                                       "np.max(", ")")

    def test_stdev(self):
        """Test the standard deviation"""
        self._check_aggregate_function(Cell.stdev, np.std, "STDEV(", ")",
                                       "np.std(", ")")

    def test_median(self):
        """Test the median"""
        self._check_aggregate_function(Cell.median, np.median, "MEDIAN(", ")",
                                       "np.median(", ")")

    def test_count(self):
        """Test the count"""
        self._check_aggregate_function(Cell.count,
                                       lambda _: self.slice_cardinality,
                                       "COUNT(", ")",
                                       "((lambda var=",
                                       ": var.shape[0] * var.shape[1])())")

    def test_irr(self):
        """Test the Internal Rate of Return (IRR)"""
        self._check_aggregate_function(Cell.irr, npf.irr, "IRR(", ")",
                                       "npf.irr(", ")")

    def test_match_negative_before_positive(self):
        """Test of finding the position of the last negative number in the
            series that is located just before the first non-negative number.
        """
        self._check_aggregate_function(Cell.match_negative_before_positive,
                                       lambda x: (np.argmin(np.array(x) < 0)),
                                       "MATCH(0,", ")",
                                       "np.argmin(", "<0)")


class TestCellUnaryFunctionality(unittest.TestCase):
    """Test unary operators (functions with just one parameter).

    Notice:
        if variable start with prefix 'a_' it means that cell is anchored
        (located in grid), if with prefix 'u_' it means that it is NOT
        anchored, with prefix 't' it does not matter (just a testing variable).
    """
    def setUp(self) -> None:
        self.cell_indices: CellIndices = CellIndices(5, 7)
        self.a_operand = Cell(3, 4, 7, cell_indices=self.cell_indices)
        self.coord_operand_python = "values[3,4]"
        self.coord_operand_excel = "F5"

        self.u_operand = Cell(value=7, cell_indices=self.cell_indices)
        self.value_u = "7"

    def test_reference(self):
        """Test the referencing to some cell"""
        with self.assertRaises(ValueError):
            Cell.reference(self.u_operand)
        u_reference = Cell.reference(self.a_operand)
        u_ref_word = u_reference.parse
        self.assertEqual(u_ref_word['excel'], "=" + self.coord_operand_excel)
        self.assertEqual(u_ref_word['python_numpy'], self.coord_operand_python)

    def test_variable(self):
        """Test the variables parsing"""
        with self.assertRaises(ValueError):
            Cell.variable(self.a_operand)

        variable_name = "test_var"
        var_value = 99
        # Un-anchored variable
        u_var_cell = Cell(None, None, 99, cell_indices=self.cell_indices,
                          is_variable=True, variable_name=variable_name,
                          cell_type=CellType.computational)
        self.assertEqual(u_var_cell.value, var_value)
        # Reference to variable
        u_ref_cell = Cell.variable(u_var_cell)
        u_ref_cell_word = u_ref_cell.parse
        self.assertEqual(u_ref_cell_word['python_numpy'], str(variable_name))
        self.assertEqual(u_ref_cell_word['excel'], '=' + str(variable_name))

    def _check_unary_operation(
            self,
            operation_method: Callable[[Cell], Cell],
            real_operation_fn: Callable[[float], float],
            excel_prefix: str = "",
            excel_suffix: str = "",
            python_prefix: str = "",
            python_suffix: str = "",
            *,
            excel_reference_prefix: str = "=",
            python_reference_prefix: str = "",
    ) -> None:
        """Run the unary operation, compare numeric result, compare words.

        Args:
            operation_method (Callable[[Cell, Cell], Cell]): Pointer to the
                method inside the Cell class.
            real_operation_fn (Callable[[float, float], float]): Pointer to the
                Python method that compute the same.
            excel_prefix (str): Prefix for the operation in Excel language.
            excel_suffix (str): Suffix for the operation in Excel language.
            python_prefix (str): Prefix for the operation in Python_NumPy
                language.
            python_suffix (str): Suffix for the operation in Python_NumPy
                language.
            excel_reference_prefix (str): Prefix of the word that reference
                to some computation in Excel.
            python_reference_prefix (str): Prefix of the word that reference
                to some computation in Python.
        """
        # A) Anchored
        a_res_cell = operation_method(self.a_operand)
        # Compare words
        a_res_parsed = a_res_cell.parse
        self.assertEqual(a_res_parsed['excel'],
                         (excel_reference_prefix + excel_prefix +
                          self.coord_operand_excel + excel_suffix)
                         )
        self.assertEqual(a_res_parsed['python_numpy'],
                         (python_reference_prefix + python_prefix +
                          self.coord_operand_python + python_suffix)
                         )
        # Compare results of anchored
        self.assertAlmostEqual(a_res_cell.value,
                               real_operation_fn(self.a_operand.value))
        # B) Un-anchored
        u_res_cell = operation_method(self.u_operand)
        # Compare words
        u_res_parsed = u_res_cell.parse
        self.assertEqual(u_res_parsed['excel'],
                         (excel_reference_prefix + excel_prefix +
                          self.value_u + excel_suffix)
                         )
        self.assertEqual(u_res_parsed['python_numpy'],
                         (python_reference_prefix + python_prefix +
                          self.value_u + python_suffix)
                         )
        # Compare results of un-anchored
        self.assertAlmostEqual(u_res_cell.value,
                               real_operation_fn(self.u_operand.value))

    def test_brackets(self):
        """Test the brackets parsing"""
        self._check_unary_operation(Cell.brackets,
                                    lambda x: x,
                                    "(", ")", "(", ")")

    def test_logarithm(self):
        """Test the logarithm parsing"""
        self._check_unary_operation(Cell.logarithm,
                                    np.log,
                                    "LN(", ")", "np.log(", ")")

    def test_exponential(self):
        """Test the exponential parsing"""
        self._check_unary_operation(Cell.exponential,
                                    np.exp,
                                    "EXP(", ")", "np.exp(", ")")

    def test_ceil(self):
        """Test the ceil parsing"""
        self._check_unary_operation(Cell.ceil,
                                    np.ceil,
                                    "CEILING(", ")", "np.ceil(", ")")

    def test_floor(self):
        """Test the floor parsing"""
        self._check_unary_operation(Cell.floor,
                                    np.floor,
                                    "FLOOR(", ")", "np.floor(", ")")

    def test_round(self):
        """Test the round parsing"""
        self._check_unary_operation(Cell.round,
                                    np.round,
                                    "ROUND(", ")", "np.round(", ")")

    def test_abs(self):
        """Test the abs parsing"""
        self._check_unary_operation(Cell.abs,
                                    np.abs,
                                    "ABS(", ")", "np.abs(", ")")

    def test_sqrt(self):
        """Test the square root parsing"""
        self._check_unary_operation(Cell.sqrt,
                                    np.sqrt,
                                    "SQRT(", ")", "np.sqrt(", ")")

    def test_logicalNegation(self):
        """Test the logical negation parsing"""
        self._check_unary_operation(Cell.logicalNegation,
                                    lambda x: not x,
                                    "NOT(", ")", "not (", ")")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="portable-spreadsheet",
    version="0.1.11",
    author="David Salac",
    author_email="info@davidsalac.eu",
    description="Simple spreadsheet that keeps tracks of each operation in "
                "defined programming languages. Logic allows export sheets to "
                "Excel files (and see how each cell is computed), to the "
                "JSON strings with description of computation e. g. in "
                "native language. Other formats like HTML, CSV and "
                "Markdown (MD) are also supported. It also allows to "
                "reconstruct behaviours in native Python with NumPy.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/david-salac/Portable-spreadsheet-generator",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'numpy-financial', 'XlsxWriter'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



from numbers import Number
from typing import Iterable, Tuple, Union, List, Optional
import copy

import numpy as np

from .cell import Cell
from .cell_indices import CellIndices
from .serialization import Serialization

# Acceptable values for the slice
T_slice = Union[np.ndarray, List[Number], List[Cell], Number, Cell]


class CellSlice(Serialization):
    """Encapsulate aggregating functionality and setting of the slices.

    Attributes:
        start_idx (Tuple[int, int]): Integer position of the starting cell
            inside the spreadsheet. Top left cell of the slice.
        end_idx (Tuple[int, int]): Integer position of the ending cell
            inside the spreadsheet. Bottom right cell of the slice.
        start_cell (Cell): Top left cell of the slice.
        end_cell (Cell): Bottom right cell of the slice.
        cell_subset (Iterable[Cell]): The list of all cells in the slice.
        driving_sheet (Spreadsheet): Reference to the spreadsheet.
    """
    def __init__(self,
                 start_idx: Tuple[int, int],
                 end_idx: Tuple[int, int],
                 cell_subset: Iterable[Cell],
                 driving_sheet
                 ):
        """Create a cell slice from the spreadsheet.

        Args:
            start_idx (Tuple[int, int]): Integer position of the starting cell
                inside the spreadsheet. Top left cell of the slice.
            end_idx (Tuple[int, int]): Integer position of the ending cell
                inside the spreadsheet. Bottom right cell of the slice.
            cell_subset (Iterable[Cell]): The list of all cells in the slice.
            driving_sheet (Spreadsheet): Reference to the spreadsheet.
        """
        # Initialise functionality for serialization:
        super().__init__(export_offset=start_idx,
                         warning_logger=driving_sheet.warning_logger,
                         export_subset=True)

        self.start_idx: Tuple[int, int] = start_idx
        self.end_idx: Tuple[int, int] = end_idx
        self.start_cell: Cell = driving_sheet.iloc[start_idx]
        self.end_cell: Cell = driving_sheet.iloc[end_idx]
        self.cell_subset: Iterable[Cell] = cell_subset
        self.driving_sheet = driving_sheet

    def sum(self, skip_none_cell: bool = True) -> Cell:
        """Compute the sum of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.sum(self.start_cell, self.end_cell, cell_subset)

    def product(self, skip_none_cell: bool = True) -> Cell:
        """Compute the product of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.product(self.start_cell, self.end_cell, cell_subset)

    def min(self, skip_none_cell: bool = True) -> Cell:
        """Find the minimum of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.min(self.start_cell, self.end_cell, cell_subset)

    def max(self, skip_none_cell: bool = True) -> Cell:
        """Find the maximum of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.max(self.start_cell, self.end_cell, cell_subset)

    def mean(self, skip_none_cell: bool = True) -> Cell:
        """Compute the mean-average of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.mean(self.start_cell, self.end_cell, cell_subset)

    def average(self, skip_none_cell: bool = True) -> Cell:
        """Compute the mean-average of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        return self.mean(skip_none_cell=skip_none_cell)

    def stdev(self, skip_none_cell: bool = True) -> Cell:
        """Compute the standard deviation of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.stdev(self.start_cell, self.end_cell, cell_subset)

    def median(self, skip_none_cell: bool = True) -> Cell:
        """Compute the median of the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.median(self.start_cell, self.end_cell, cell_subset)

    def count(self, skip_none_cell: bool = True) -> Cell:
        """Compute the number of items in the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.count(self.start_cell, self.end_cell, cell_subset)

    def irr(self, skip_none_cell: bool = True) -> Cell:
        """Compute the Internal Rate of Return (IRR) of items in the aggregate.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: a new cell with the result.
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.irr(self.start_cell, self.end_cell, cell_subset)

    def match_negative_before_positive(self,
                                       skip_none_cell: bool = True) -> Cell:
        """Find the position of the last negative number in the series that
            is located just before the first non-negative number.

        Args:
            skip_none_cell (bool): If true, skips all the cells with None as
                a value (and does not raise exception).

        Returns:
            Cell: Return the position of the negative number in a series that
                is located just before the first positive number (or zero).
        """
        if skip_none_cell:
            cell_subset = [nn_cell for nn_cell in self.cell_subset
                           if nn_cell.value is not None]
        else:
            cell_subset = self.cell_subset
        return Cell.match_negative_before_positive(self.start_cell,
                                                   self.end_cell, cell_subset)

    @property
    def excel_format(self):
        """Should not be accessible for slides."""
        raise NotImplementedError

    @excel_format.setter
    def excel_format(self, new_format: dict):
        """Set the Excel cell format/style.

        Read the documentation: https://xlsxwriter.readthedocs.io/format.html

        Args:
            new_format (dict): New format definition.
        """
        if not isinstance(new_format, dict):
            raise ValueError("New format has to be a dictionary!")
        for row in range(self.start_idx[0], self.end_idx[0] + 1):
            for col in range(self.start_idx[1],
                             self.end_idx[1] + 1):
                self.driving_sheet.iloc[row, col].excel_format = new_format

    @property
    def description(self) -> Optional[str]:
        """Not implementable.
        """
        raise NotImplementedError

    @description.setter
    def description(self, new_description: Optional[str]):
        """Set the cell description.

        Args:
            new_description (Optional[str]): description of the cell.
        """
        if (new_description is not None
                and not isinstance(new_description, str)):
            raise ValueError("Cell description has to be a string value!")
        for row in range(self.start_idx[0], self.end_idx[0] + 1):
            for col in range(self.start_idx[1],
                             self.end_idx[1] + 1):
                self.driving_sheet.iloc[row, col].description = new_description

    def _set_value_on_position(self, other: Union[Cell, Number],
                               row: int, col: int) -> None:
        """Set the cell on given position in the spreadsheet to the value
            'other'.
        Args:
            other (Union[Cell, Number]): new value to be set.
            row (int): the row integer position in the spreadsheet (indexed
                from 0).
            col (int): the column integer position in the spreadsheet (indexed
                from 0).
        """
        if isinstance(other, Cell):
            # Set the right values
            if other.anchored:
                self.driving_sheet.iloc[row, col] = \
                    Cell.reference(other)
            else:
                # Create a deep copy
                self.driving_sheet.iloc[row, col] = \
                    copy.deepcopy(other)
                # Anchor it:
                self.driving_sheet.iloc[row, col].coordinates = (row, col)
        else:
            # Call the external logic to manage the same
            self.driving_sheet.iloc[row, col] = other

    # Set to scalar / Other cells:
    def set(self, other: T_slice) -> None:
        """Set all the values in the slice to the new one (or the list of
            values).

        Args:
            other: Union[np.ndarray, List[Number], List[Cell], Number, Cell]:
                Some value or list (or numpy array) of values that should be
                set for all the cells inside slice.
        """
        if isinstance(other, (np.ndarray, list)):
            dim_match = True
            is_list = True
            is_1d = False
            by_row = self.shape[0] > self.shape[1]
            if hasattr(other, "shape"):
                dim_match = other.shape == self.shape
                is_list = False
                is_1d = len(other.shape) == 1
                if is_1d:
                    dim_match = max(other.shape) == max(self.shape)
            else:
                is_list = True
                if min(self.shape) == 1:
                    dim_match = len(other) == max(self.shape)
                    is_1d = True
            if not dim_match:
                raise ValueError("Shape of the input does not match to the "
                                 "shape of the slice!")
            if is_1d:
                col = self.start_idx[1]
                row = self.start_idx[0]
                for val in other:
                    self._set_value_on_position(val, row, col)
                    if by_row:
                        row += 1
                    else:
                        col += 1
            else:
                # If is N-dimensional
                for row in range(self.start_idx[0], self.end_idx[0] + 1):
                    for col in range(self.start_idx[1],
                                     self.end_idx[1] + 1):
                        if is_list:
                            val = other[row - self.start_idx[0]][
                                col - self.start_idx[1]
                            ]
                        else:
                            val = other[
                                row - self.start_idx[0],
                                col - self.start_idx[1]
                            ]
                        self._set_value_on_position(val, row, col)

        else:
            for row in range(self.start_idx[0], self.end_idx[0] + 1):
                for col in range(self.start_idx[1], self.end_idx[1] + 1):
                    # Set the right values
                    self._set_value_on_position(other, row, col)

    def __ilshift__(self, other: T_slice):
        """Overrides operator <<= to do a set functionality.
        """
        self.set(other)

    # ==== OVERRIDE ABSTRACT METHODS AND PROPERTIES OF SERIALIZATION CLASS ====
    @Serialization.shape.getter
    def shape(self) -> Tuple[int, int]:
        """Return the shape of the sheet in the NumPy logic.

        Returns:
            Tuple[int]: Number of rows, Number of columns
        """
        return (self.end_idx[0] - self.start_idx[0] + 1,
                self.end_idx[1] - self.start_idx[1] + 1)

    @Serialization.cell_indices.getter
    def cell_indices(self) -> CellIndices:
        """Get the cell indices.

        Returns:
            CellIndices: Cell indices of the spreadsheet.
        """
        return self.driving_sheet._cell_indices

    def _get_cell_at(self, row: int, column: int) -> 'Cell':
        """Get the particular cell on the (row, column) position.

        Returns:
            Cell: The call on given position.
        """
        return self.driving_sheet.iloc[self.start_idx[0] + row,
                                       self.start_idx[1] + column]

    def _get_variables(self):
        """Return the sheet variables as _SheetVariables object.

        Returns:
            _SheetVariables: Sheet variables.
        """
        return self.driving_sheet.var
    # =========================================================================
